import functools
from abc import ABC, abstractmethod
from typing import Set, Tuple, Type

from importlib_metadata import EntryPoint, entry_points

from defederate.const import BlockDef


class ServerPlugin(ABC):
    @abstractmethod
    def __init__(self, server: str):
        ...

    def get_public_blocklist(self) -> Set[BlockDef]:
        raise NotImplementedError("This function is not supported by this server type")

    @staticmethod
    def understands(url: str) -> bool:
        return False


@functools.lru_cache()
def list_server_plugins() -> Tuple[EntryPoint]:
    registered_plugins = entry_points(group="defederate.plugin.server")
    if not registered_plugins:
        raise RuntimeError(
            "Entry points missing. Defederation package is not installed correctly."
        )
    return registered_plugins


def get_server_plugin(name: str) -> Type[ServerPlugin]:
    plugin = [ep for ep in list_server_plugins() if ep.name == name]
    if not plugin:
        raise KeyError(f"Could not find plugin {name}")
    if len(plugin) > 1:
        raise KeyError(f"Plugin {name} is ambiguous")
    entry_point = plugin[0].load()
    if not issubclass(entry_point, ServerPlugin):
        raise TypeError(
            f"Plugin {name} is invalid - returned {entry_point}, not ServerPlugin object"
        )
    return entry_point



import functools
import json
import urllib.parse
from typing import Optional

import requests


@functools.lru_cache(maxsize=None)
def get_nodeinfo(url: str) -> dict:
    parsed_url = urllib.parse.urlparse(url)
    well_known_nodeinfo = parsed_url._replace(
        path="/.well-known/nodeinfo", params="", query="", fragment=""
    )
    ap_nodeinfo = requests.get(well_known_nodeinfo.geturl())
    if ap_nodeinfo.status_code != 200:
        return {}
    ap_nodeinfo_data = json.loads(ap_nodeinfo.text)
    nodeinfo_url: Optional[str] = None
    for entry in ap_nodeinfo_data.get("links", []):
        if entry.get("rel") == "http://nodeinfo.diaspora.software/ns/schema/2.0":
            nodeinfo_url = entry.get("href")
            break
    if not nodeinfo_url:
        return {}

    nodeinfo = requests.get(nodeinfo_url)
    if nodeinfo.status_code != 200:
        return {}
    nodeinfo_data = json.loads(nodeinfo.text)
    if not isinstance(nodeinfo_data, dict):
        return {}
    return nodeinfo_data



import enum
from typing import NamedTuple, Optional


class BlockLevel(enum.IntEnum):
    NONE = enum.auto()
    SILENCE = enum.auto()
    SUSPEND = enum.auto()


class BlockSubmission(NamedTuple):
    domain: str
    severity: BlockLevel
    reject_media: bool = False
    reject_reports: bool = False
    obfuscate: bool = False
    private_comment: str = ""
    public_comment: str = ""


class BlockDef(NamedTuple):
    source: str
    server: str
    level: BlockLevel
    reason: str
    digest: Optional[str] = None



import json
import urllib.parse
from typing import Set

import requests

from defederate.activitypub import get_nodeinfo
from defederate.const import BlockDef, BlockLevel
from defederate.errors import CannotScrapePage
from defederate.plugin import ServerPlugin


class Mastodon4Server(ServerPlugin):
    def __init__(self, url: str):
        self._server = urllib.parse.urlparse(url)

    @staticmethod
    def understands(url: str) -> bool:
        parsed = urllib.parse.urlparse(url)
        if parsed.path not in ("", "/"):
            return False
        nodeinfo = get_nodeinfo(url)
        if nodeinfo.get("software", {}).get("name") != "mastodon":
            return False
        return nodeinfo["software"].get("version", "").startswith("4.")

    def get_public_blocklist(self) -> Set[BlockDef]:
        url = self._server._replace(
            path="/api/v1/instance/domain_blocks", params="", query="", fragment=""
        ).geturl()
        response = requests.get(url)
        if response.status_code != 200:
            raise CannotScrapePage(
                f"Server at {self._server.netloc} returned {response.status_code}: {response.reason}"
            )
        levels = {"suspend": BlockLevel.SUSPEND, "silence": BlockLevel.SILENCE}
        return {
            BlockDef(
                source=self._server.netloc,
                server=entry["domain"],
                digest=entry["digest"],
                level=levels.get(entry["severity"], BlockLevel.NONE),
                reason=entry["comment"] or "",
            )
            for entry in json.loads(response.text)
        }



import collections
import hashlib
import json
import os.path
import re
import urllib.request
from typing import Dict, Optional, Set

from .const import BlockDef, BlockLevel

OFFLINE = os.path.exists("mymasto")


def sha256(text: str) -> str:
    m = hashlib.sha256()
    m.update(text.encode("utf8"))
    return m.hexdigest()


def get(url: str) -> str:
    response = urllib.request.urlopen(url)
    data = response.read()
    text = data.decode("utf-8")
    return text


def parse_markdown(source: str, url: str) -> Set[BlockDef]:
    parse = re.compile(
        r"^\|(?P<server>[^|]+)\|\s*(?P<mark>[^ |]+)\s*\|\s*(?P<reason>[^|]*)\|",
        re.MULTILINE,
    )
    if OFFLINE:
        data = open("data").read()
    else:
        data = get(url)
    blocks = set()
    levels = collections.defaultdict(lambda: BlockLevel.NONE)
    levels["\N{NO ENTRY}"] = BlockLevel.SUSPEND
    levels["\N{SPEAKER WITH CANCELLATION STROKE}"] = BlockLevel.SILENCE
    for line in data.split("\n"):
        definition = parse.match(line)
        if definition:
            servers, mark, reason = definition.groups()
            for server in servers.split(","):
                if "." not in server:
                    continue
                blocks.add(
                    BlockDef(
                        source=source,
                        server=server.strip(),
                        level=levels[mark],
                        reason=reason.strip(),
                    )
                )
    return blocks


def parse_mastodon4(server: str, url: Optional[str] = None) -> Set[BlockDef]:
    url = url or f"https://{server}/api/v1/instance/domain_blocks"
    if OFFLINE:
        datastr = open("masto").read()
    else:
        datastr = get(url)
    data = set()
    levels = collections.defaultdict(lambda: BlockLevel.NONE)
    levels["suspend"] = BlockLevel.SUSPEND
    levels["silence"] = BlockLevel.SILENCE
    for entry in json.loads(datastr):
        data.add(
            BlockDef(
                source=server,
                server=entry["domain"],
                digest=entry["digest"],
                level=levels[entry["severity"]],
                reason=entry["comment"] or "",
            )
        )
    return data


def parse_mastodon3(server: str, url: Optional[str] = None) -> Set[BlockDef]:
    url = url or f"https://{server}/about/more"
    results: Set[BlockDef] = set()

    if OFFLINE:
        datastr = open("mymasto").read()
    else:
        datastr = get(url)
    start = re.split(r"<h. id='unavailable-content'>", datastr, maxsplit=1)
    if len(start) < 2:
        return results
    datastr = start[1]

    limited = re.search(
        r"<h.>Limited servers(.*?)</table>", datastr, re.IGNORECASE | re.DOTALL
    )
    suspended = re.search(
        r"<h.>Suspended servers(.*?)</table>", datastr, re.IGNORECASE | re.DOTALL
    )
    blockexpr = re.compile(
        r"<td\s*class[^>]*>\s*<span(?: title='SHA-256: (?P<hash>[a-fA-F0-9]*)')?[^>]*>"
        r"(?P<server>[^<]*)</span>\s*</td>\s*<td[^>]*>(?P<reason>[^<]*)</td>(?P<optional>asdf)?",
        re.IGNORECASE | re.DOTALL,
    )

    def parse_and_add(result_iterator, at_level):
        for entry in result_iterator:
            results.add(
                BlockDef(
                    source=server,
                    server=entry["server"],
                    digest=entry["hash"],
                    level=at_level,
                    reason=entry["reason"].strip(),
                )
            )

    if limited:
        parse_and_add(blockexpr.finditer(limited.group(1)), BlockLevel.SILENCE)
    if suspended:
        parse_and_add(blockexpr.finditer(suspended.group(1)), BlockLevel.SUSPEND)

    return results


if __name__ == "__main__":
    remote_sources = [
        parse_markdown(
            "chaos.social",
            "https://raw.githubusercontent.com/chaossocial/about/07c5d774157b847a796b8b126990277ee1643b66/blocked_instances.md",
        ),
        parse_mastodon4("mas.to"),
    ]
    local_definition = parse_mastodon3("mast.uxp.de")
    exceptions = {
        "birdsite.thorlaksson.com": BlockLevel.SILENCE,  # Twitter-xpost is fine
        "shitpost.cloud": BlockLevel.SILENCE,
    }

    blocklist_remote: Dict[str, BlockDef] = {}
    blocklist_local: Dict[str, BlockDef] = {}
    for source in remote_sources:
        for entry in source:
            digest = entry.digest or sha256(entry.server)
            if digest in blocklist_remote:
                # identify the server name if possible
                if "*" in blocklist_remote[digest].server and "*" not in entry.server:
                    blocklist_remote[digest] = blocklist_remote[digest]._replace(
                        server=entry.server
                    )
                elif "*" not in blocklist_remote[digest].server and "*" in entry.server:
                    entry = entry._replace(server=blocklist_remote[digest].server)
                if blocklist_remote[digest].level >= entry.level:
                    continue  # highest level wins, earlier entry has priority
            if "*" in entry.server and not entry.digest:
                continue  # can't block this
            blocklist_remote[digest] = entry
    for server, level in exceptions.items():
        digest = sha256(server)
        blocklist_remote[digest] = BlockDef(
            source="exception",
            server=server,
            level=level,
            digest=digest,
            reason="manual exception",
        )
    for entry in local_definition:
        if "*" in entry.server and not entry.digest:
            continue  # can't parse this
        digest = entry.digest or sha256(entry.server)
        if digest in blocklist_local and blocklist_local[digest].level <= entry.level:
            continue  # highest level wins
        blocklist_local[digest] = entry

    unseen = blocklist_remote.keys() - blocklist_local.keys()
    unknown = blocklist_local.keys() - blocklist_remote.keys()
    known = blocklist_local.keys() & blocklist_remote.keys()

    if unseen:
        print("\nThe following servers need adding to the blocklist:")
        reasons = sorted(
            "* {server.server} listed on {server.source} as {server.level.name}{dueto}".format(
                server=blocklist_remote[digest],
                dueto=" due to " + blocklist_remote[digest].reason
                if blocklist_remote[digest].reason
                else "",
            )
            for digest in unseen
            if "*" not in blocklist_remote[digest].server
        )
        print("\n".join(reasons))

    if known:
        divergent = sorted(
            (
                "* {local.server} listed locally as {local.level.name} "
                "but is reported as {remote.level.name} on {remote.source}{dueto}"
            ).format(
                local=blocklist_local[digest],
                remote=blocklist_remote[digest],
                dueto=" due to " + blocklist_remote[digest].reason
                if blocklist_remote[digest].reason
                else "",
            )
            for digest in known
            if blocklist_local[digest].level is not blocklist_remote[digest].level
        )
        if divergent:
            print("\nThe following servers need updating:")
            print("\n".join(divergent))

    if unknown:
        print("\nThe following servers are no longer listed and can be removed:")
        reasons = sorted(
            "* {server.server} listed locally as {server.level.name}{dueto}".format(
                server=blocklist_local[digest],
                dueto=" due to " + blocklist_local[digest].reason
                if blocklist_local[digest].reason
                else "",
            )
            for digest in unknown
        )
        print("\n".join(reasons))

    print()



import argparse
import sys
import urllib.parse
from typing import List

import defederate
import defederate.blocklist_parsers
import defederate.plugin


def _sanitize_address(url: str) -> str:
    parsed = urllib.parse.urlparse(url)
    if not parsed.scheme and not parsed.netloc:
        parsed = urllib.parse.urlparse("https://" + url)
    return parsed.geturl()


def main():
    parser = argparse.ArgumentParser(
        usage="defederate <command> [<args>]",
        description="""Tools for Mastodon Instance Blocklist Management (v{version})
The most commonly used commands are:
   show     Show currently active instance blocks
   block    Add an instance block
Each command has its own set of parameters, and you can get more information
by running defederate <command> --help
""".format(
            version=defederate.__version__
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("subcommand", help=argparse.SUPPRESS)
    # parse_args defaults to [1:] for args, but need to
    # exclude the rest of the args too, or validation will fail
    parameters = sys.argv[1:2]
    if not parameters:
        parser.print_help()
        sys.exit(0)
    args = parser.parse_args(parameters)
    subcommand = globals().get("cli_" + args.subcommand)
    if subcommand:
        return subcommand(sys.argv[2:])
    parser.print_help()
    print()
    sys.exit(f"Unrecognized command: {args.subcommand}")


def cli_show(cmd_args: List[str]):
    known_server_types: List[str] = sorted(
        ep.name for ep in defederate.plugin.list_server_plugins()
    )
    parser = argparse.ArgumentParser(
        description="Show active instance blocks", prog="defederate show"
    )
    parser.add_argument(
        "--mastodon-version",
        choices=known_server_types,
        help="Specify the version of Mastodon running on this server",
    )
    parser.add_argument(
        "host", help="The base URL for the server", type=_sanitize_address
    )

    args = parser.parse_args(cmd_args)
    if not args.mastodon_version:
        accepting_plugins = sorted(
            server_type
            for server_type in known_server_types
            if defederate.plugin.get_server_plugin(server_type).understands(args.host)
        )
        if not accepting_plugins:
            exit(
                "Could not identify the server type. Please specify --mastodon-version"
            )
        if len(accepting_plugins) > 1:
            exit(
                f"Could not identify the server type. Please specify --mastodon-version with one of {accepting_plugins}"
            )
        args.mastodon_version = accepting_plugins[0]
        print(f"Autodetected server type {args.mastodon_version}")

    server = defederate.plugin.get_server_plugin(args.mastodon_version)(args.host)
    blockset = server.get_public_blocklist()
    blocklist = sorted(
        " {entry.level.name}: {entry.server}{due}".format(
            entry=entry, due=f"  due to {entry.reason}" if entry.reason else ""
        )
        for entry in blockset
    )
    print(f"Current blocklist on {args.host}:")
    print("\n".join(blocklist))


def cli_block(args):
    exit("Adding blocks is not yet supported")



import re
import urllib.parse
from typing import Set

import requests

from defederate.const import BlockDef, BlockLevel
from defederate.errors import CannotScrapePage
from defederate.plugin import ServerPlugin

_table_parser = re.compile(
    r"^\|(?P<server>[^|]+)\|\s*(?P<mark>[^ |]+)\s*\|\s*(?P<reason>[^|]*)\|",
    re.MULTILINE,
)


class MarkdownServer(ServerPlugin):
    def __init__(self, server: str):
        self._server = server

    @staticmethod
    def understands(url: str) -> bool:
        parsed = urllib.parse.urlparse(url)
        return parsed.path[-3:].upper() == ".MD"

    def get_public_blocklist(self) -> Set[BlockDef]:
        url = self._server
        response = requests.get(url)
        if response.status_code != 200:
            raise CannotScrapePage(
                f"Server at {url} returned {response.status_code}: {response.reason}"
            )
        levels = {
            "\N{NO ENTRY}": BlockLevel.SUSPEND,
            "\N{SPEAKER WITH CANCELLATION STROKE}": BlockLevel.SILENCE,
        }

        blocks: Set[BlockDef] = set()
        for line in response.text.split("\n"):
            definition = _table_parser.match(line)
            if definition:
                servers, mark, reason = definition.groups()
                for server in servers.split(","):
                    if "." not in server:
                        continue
                    blocks.add(
                        BlockDef(
                            source=self._server,
                            server=server.strip(),
                            level=levels.get(mark, BlockLevel.NONE),
                            reason=reason.strip(),
                        )
                    )
        return blocks



import re
import urllib.parse
from typing import Set

import requests

from defederate.activitypub import get_nodeinfo
from defederate.const import BlockDef, BlockLevel
from defederate.errors import CannotScrapePage
from defederate.plugin import ServerPlugin


class Mastodon3Server(ServerPlugin):
    def __init__(self, server: str):
        self._server = urllib.parse.urlparse(server)

    @staticmethod
    def understands(url: str) -> bool:
        parsed = urllib.parse.urlparse(url)
        if parsed.path not in ("", "/"):
            return False
        nodeinfo = get_nodeinfo(url)
        if nodeinfo.get("software", {}).get("name") != "mastodon":
            return False
        return nodeinfo["software"].get("version", "").startswith("3.")

    def get_public_blocklist(self) -> Set[BlockDef]:
        url = self._server._replace(
            path="/about/more", params="", query="", fragment=""
        ).geturl()
        response = requests.get(url)
        if response.status_code != 200:
            raise CannotScrapePage(
                f"Server at {self._server.netloc} returned {response.status_code}: {response.reason}"
            )

        start = re.split(r"<h. id='unavailable-content'>", response.text, maxsplit=1)
        if len(start) < 2:
            raise CannotScrapePage(
                f"Server at {self._server.netloc} returned page in unexpected format"
            )
        server_list = start[1]
        results = set()

        limited = re.search(
            r"<h.>Limited servers(.*?)</table>", server_list, re.IGNORECASE | re.DOTALL
        )
        suspended = re.search(
            r"<h.>Suspended servers(.*?)</table>",
            server_list,
            re.IGNORECASE | re.DOTALL,
        )
        blockexpr = re.compile(
            r"<td\s*class[^>]*>\s*<span(?: title='SHA-256: (?P<hash>[a-fA-F0-9]*)')?[^>]*>"
            r"(?P<server>[^<]*)</span>\s*</td>\s*<td[^>]*>(?P<reason>[^<]*)</td>(?P<optional>asdf)?",
            re.IGNORECASE | re.DOTALL,
        )

        def parse_and_add(result_iterator, at_level):
            for entry in result_iterator:
                results.add(
                    BlockDef(
                        source=self._server.netloc,
                        server=entry["server"],
                        digest=entry["hash"],
                        level=at_level,
                        reason=entry["reason"].strip(),
                    )
                )

        if limited:
            parse_and_add(blockexpr.finditer(limited.group(1)), BlockLevel.SILENCE)
        if suspended:
            parse_and_add(blockexpr.finditer(suspended.group(1)), BlockLevel.SUSPEND)

        return results



import re
import urllib.parse

import requests

from .const import BlockLevel, BlockSubmission
from .errors import CannotScrapePage


def cookie_decode(input: str) -> str:
    if "%" in input:
        return urllib.parse.unquote(input)
    return input


class Mastodon3Web:
    def __init__(self, server_url, mastodon_session, session_id):
        self._mastodon_session = cookie_decode(mastodon_session)
        self._session_id = cookie_decode(session_id)
        self._server_url = server_url.rstrip("/")

    def add(self, block: BlockSubmission):
        cookies = {
            "_mastodon_session": self._mastodon_session,
            "_session_id": self._session_id,
        }
        headers = {"User-Agent": "Mozilla/5.0", "Accept_Language": "en-US,en;q=0.5"}
        new_page = f"{self._server_url}/admin/domain_blocks/new"
        response = requests.get(new_page, cookies=cookies, headers=headers)

        csrf = re.search(
            r'<meta name="csrf-token" content="([^"]*)"', response.text, re.DOTALL
        )
        if not csrf:
            raise CannotScrapePage("Could not find CSRF token on domain block page")
        headers["X-CSRF-Token"] = csrf.group(1)

        form = re.search(r"(<form(.*?)</form>)", response.text, re.DOTALL)
        if not form:
            raise CannotScrapePage("Could not identify form on domain block page")
        auth_match = re.search(
            r'<input [^>]*name="authenticity_token"[^>]*value="([^"]+)"',
            form.group(1),
            re.DOTALL,
        )
        if not auth_match:
            raise CannotScrapePage("Could not find authentication token on page")

        payload = {
            "authenticity_token": auth_match.group(1),
            "domain_block[domain]": block.domain,
            "domain_block[reject_media]": "1" if block.reject_media else "0",
            "domain_block[reject_reports]": "1" if block.reject_reports else "0",
            "domain_block[obfuscate]": "1" if block.obfuscate else "0",
            "domain_block[private_comment]": block.private_comment,
            "domain_block[public_comment]": block.public_comment,
            "button": "",
        }
        if block.severity is BlockLevel.SUSPEND:
            payload["domain_block[severity]"] = "suspend"
        elif block.severity is BlockLevel.SILENCE:
            payload["domain_block[severity]"] = "silence"
        else:
            payload["domain_block[severity]"] = "noop"

        block_response = requests.post(
            f"{self._server_url}/admin/domain_blocks",
            cookies=response.cookies,
            data=payload,
            headers=headers,
        )
        if block_response.status_code == 200:
            return
        raise CannotScrapePage(
            f"Server returned {block_response.status_code}: {block_response.reason}"
        )


mw = Mastodon3Web(
    "https://mast.uxp.de",
    "mastodon session goes here",
    "session id goes here",
)

mw.add(
    BlockSubmission(
        "beefyboys.club",
        severity=BlockLevel.SUSPEND,
        public_comment="test suspension :)",
    )
)
